/*****************************************************************************/
/*	ascd - AfterStep CD 				             */
/*	Version 0.5						             */
/*	By Rob Malda							     */		
/*	malda@cs.hope.edu						     */
/*	http://www.cs.hope.edu/~malda/					     */
/*									     */
/*	This is an 'AfterStep Look & Feel' Wharf style applet that can be    */
/*	used to control an IDE CD Changer.				     */
/*									     */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>       
#include <string.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#include <errno.h>
#include <unistd.h>
#include <linux/cdrom.h>
#include <linux/ucdrom.h>
/* #include <linux/fs.h>   */                         

#include <sys/mount.h>
#include "hardware.c"

  
/* XPM struct and icons ******************************************************/
typedef struct _XpmIcon {
	Pixmap pixmap;
	Pixmap mask;
	XpmAttributes attributes;
} XpmIcon;

XpmIcon cdplayerXPM,  ledXPM, cdpausedXPM, currentXPM;

#include "XPM/led.xpm"
#include "XPM/cdplayer.xpm"
#include "XPM/cdpaused.xpm"
/* Functions *****************************************************************/
void	Help(void);
void	CreateWindow(int, char **);
void	ParseCmdLine(int argc, char *argv[]);
void	MainLoop();
void	GetXPM(void);
int	FlushExpose(Window w);
void	RedrawWindow( XpmIcon *v);
Pixel	GetColor(char *name);
void	cdcontrol(int control);



/* Global stuff **************************************************************/
/* #define TRUE 1; */
/* #define FALSE 0; */
#define DEFAULTDEVICE "/dev/hdd";

#define PLAY 0
#define PAUSE 1
#define STOP 2
#define UPTRACK 3
#define DNTRACK 4

int	lasttime=-1;
Display	*Disp;	 
Window	Root;
Window	Iconwin;
Window	Win;
char	*Geometry = 0;
char    device[128]=DEFAULTDEVICE;
int	withdrawn=FALSE;

int 	currenttrack=0;
int 	maxtracks=31;

int  	paused=FALSE;

int 	playing=FALSE;
GC	WinGC;
int     CarrierOn = FALSE;

/*****************************************************************************/
int main(int argc,char *argv[])
{       	
	ParseCmdLine(argc, argv);   
	cd_status();
          /* Olly patch */
          switch (cur_cdmode) {
            case CDPLAY: playing=TRUE; break;
            case CDPAUZ: paused=TRUE; break;
          }                                   
	maxtracks=thiscd.ntracks+1; 
	close(cd_fd); 
	cd_fd=-1;
        CreateWindow(argc, argv);
        XSetCommand(Disp,Win,argv,argc);    
	MainLoop();
	return 0;
}







void cdcontrol(int order)
{
        int status=0;       /* return status for system calls */
        /* this variable keeps is TRUE if the track is changed whilst the
        cd is paused, if it was changed, then start from "0" otherwise
        start from the last known position (which is gotten from cd_status() */
        static int pos_changed = FALSE;
	if(cd_fd < 0) cd_fd=open(device,0);
	cd_status();
	lasttime=-1;
	switch(order)
                        {
                         case PLAY: cd_status();
				    maxtracks= thiscd.ntracks+1;
				    if(currenttrack==0) currenttrack=1;
				    if(paused)
					{ paused=TRUE; }
				    else
					{ 
					play_cd(currenttrack,0,maxtracks);	
					}
			            playing=TRUE;
				    paused=FALSE;  
                                break;
                         case PAUSE:
				    if(playing)
					{ 
					cur_cdmode=1;
				    	pause_cd();
				    	paused=TRUE;
				    	playing=FALSE;
					}
				    else if(paused)
					{
					cur_cdmode=3;
					cd_status();
					if(pos_changed)
						play_cd(currenttrack, 0, maxtracks);
					else
						play_cd(currenttrack,cur_pos_rel,maxtracks);
					/*pause_cd();*/
					paused=FALSE;
					pos_changed = FALSE;
					playing=TRUE;
					}  
                                break;
			  case STOP:
				    if(playing)
					 { 
					   currenttrack=0; 
					   cur_cdmode=3;
					   stop_cd();
					 }
				    else { currenttrack=0;
					   eject_cd(); 
					  }
		                    currenttrack=0; 
			            playing=FALSE;
				    paused=FALSE;
				    pos_changed = FALSE;
				    close(cd_fd);
				    cd_fd=-1; 
                                break;
                         case UPTRACK:  
                    if(paused)
             			{
             			cur_pos_rel = 0;
             			pos_changed = TRUE;
             			} 
				    currenttrack++;
				    if(currenttrack >= maxtracks) currenttrack=1;
				    if(playing) play_cd(currenttrack,0,maxtracks);
				    /* If Playing, Stop, Uptrack, Play */
				    /* If Paused  */
                                break;
			 case DNTRACK:
		        cd_status();
			 	if(paused)
			 		{
			 		if(cur_pos_rel < 2 || pos_changed)
				 		currenttrack--;
			 		cur_pos_rel = 0;
			 		pos_changed = TRUE;
			 		}
				else if(cur_pos_rel < 2)
		                    currenttrack--;
				    
				    if(currenttrack < 1) currenttrack=maxtracks-1; 
				    if(playing) play_cd(currenttrack,0,maxtracks);
				break;
                        }                   

        RedrawWindow(&currentXPM); 	
        

}                          







/*****************************************************************************/
void Help()
{       
	fprintf(stderr,"ascd - Version 0.5\n");
	fprintf(stderr,"by Rob Malda - malda@cs.hope.edu\n");
	fprintf(stderr,"http://www.cs.hope.edu/~malda/\n");
	fprintf(stderr,"usage:  ascd [-options ...] \n");
	fprintf(stderr,"options:\n");
	fprintf(stderr,"\n");       
	fprintf(stderr,"  Command                 Default   Why\n");
	fprintf(stderr,"g Geometry                none      Standard X Location\n");	
	fprintf(stderr,"d Device                  /dev/hdd  IDE Device of the Drive\n");
	fprintf(stderr,"w Withdrawn		  none	    Withdrawn more (for WindowMaker)\n");
	fprintf(stderr,"\n\n");
	exit(1); 
}

/****************************************************************************/
void CreateWindow(int argc, char **argv)
{
	int i;
	unsigned int borderwidth ;
	char *display_name = NULL; 
	char *wname = "ascd";
	XGCValues gcv;
	unsigned long gcm;
	XTextProperty name;
	Pixel back_pix, fore_pix;
	Pixmap pixmask;
	int screen;	
	int x_fd;
	int d_depth;
	int ScreenWidth, ScreenHeight;
	XSizeHints SizeHints;
	XWMHints WmHints;
	XClassHint classHint;
	
	/* Open display */
	if (!(Disp = XOpenDisplay(display_name)))  
	{ 
		fprintf(stderr,"ascd: can't open display %s\n", XDisplayName(display_name));    
		exit (1); 
	} 
	
	screen = DefaultScreen(Disp);
	Root = RootWindow(Disp, screen);
	d_depth = DefaultDepth(Disp, screen);
	x_fd = XConnectionNumber(Disp);	
	ScreenHeight = DisplayHeight(Disp,screen);
	ScreenWidth = DisplayWidth(Disp,screen);
	       
	GetXPM();
		
	SizeHints.flags= USSize|USPosition;
	SizeHints.x = 0;
	SizeHints.y = 0;	
	back_pix = GetColor("white");
	fore_pix = GetColor("black");
	
	XWMGeometry(Disp, screen, Geometry, NULL, (borderwidth =1), &SizeHints,
		    &SizeHints.x,&SizeHints.y,&SizeHints.width,
		    &SizeHints.height, &i); 	
	
	SizeHints.width = cdplayerXPM.attributes.width;
	SizeHints.height= cdplayerXPM.attributes.height;	
	Win = XCreateSimpleWindow(Disp,Root,SizeHints.x,SizeHints.y,
				  SizeHints.width,SizeHints.height,
				  borderwidth,fore_pix,back_pix);
	Iconwin = XCreateSimpleWindow(Disp,Win,SizeHints.x,SizeHints.y,
				      SizeHints.width,SizeHints.height,
				      borderwidth,fore_pix,back_pix);	     
	XSetWMNormalHints(Disp, Win, &SizeHints);
	
	classHint.res_name = "ascd";
	classHint.res_class = "AScd";
	XSetClassHint (Disp, Win, &classHint);
	
	XSelectInput(Disp, Win, (ExposureMask | ButtonPressMask | 
				 StructureNotifyMask));
	XSelectInput(Disp, Iconwin, (ExposureMask | ButtonPressMask | 
				     StructureNotifyMask));
	
	if (XStringListToTextProperty(&wname, 1, &name) ==0) 
	{
		fprintf(stderr, "ascd: can't allocate window name\n");
		exit(-1);
	}
	XSetWMName(Disp, Win, &name);
	
	/* Create WinGC */
	gcm = GCForeground|GCBackground|GCGraphicsExposures;
	gcv.foreground = fore_pix;
	gcv.background = back_pix;
	gcv.graphics_exposures = False;
	WinGC = XCreateGC(Disp, Root, gcm, &gcv);  

	 
		
		
	WmHints.initial_state = withdrawn ? WithdrawnState : NormalState;
	WmHints.icon_window = Iconwin;
	WmHints.window_group = Win;
	WmHints.flags = StateHint | IconWindowHint | IconPositionHint
		| WindowGroupHint;
	WmHints.icon_x = SizeHints.x;
	WmHints.icon_y = SizeHints.y;
	XSetWMHints(Disp, Win, &WmHints); 	
	XSetCommand (Disp, Win, argv, argc);
	XMapWindow(Disp,Win);
	RedrawWindow(&currentXPM);
}

/****************************************************************************/
void ParseCmdLine(int argc, char *argv[])
{

	char *Argument;
	int i;       
	
	for(i = 1; i < argc; i++) 
	{
		Argument = argv[i];
		
		if (Argument[0] == '-') 
		{
			switch(Argument[1]) 
			{
                         case 'w':
           			withdrawn=TRUE;
				continue;
			 case 'g': /* Geometry */
				if(++i >= argc) Help();
				Geometry = argv[i];
				continue;
			 case 'd': /* Device */
				if(++i >= argc) Help();
				strcpy(&device[0], argv[i]);
				continue;
			case 'h':  /* Help */
				if(++i >= argc) Help();
				continue;
			
			 default:
				Help();
			}
		}
		else
			Help();
	}

}

/****************************************************************************/
void MainLoop()
{
	XEvent Event;            
			      
	/* Main loop */
	while(1)
	{		
		/* Check events */
		while (XPending(Disp))
		{
			XNextEvent(Disp,&Event);
			switch(Event.type)
			{		     
			 case Expose:		/* Redraw window */
				if(Event.xexpose.count == 0)
					{ lasttime=01; RedrawWindow(&currentXPM); } 
				break;		     
			 case ButtonPress:	/* Mouseclick */
				
				if(Event.xbutton.y < 15)
                                    {
				      if(Event.xbutton.x > 24) cdcontrol(UPTRACK);
				    }
				else if(Event.xbutton.y < 32)
				    { if(Event.xbutton.x < 24) 
					{if(playing || paused) 
                                           cdcontrol(PAUSE);
				         else
					   cdcontrol(PLAY);
				        }
				      else 
			                { cdcontrol(STOP);

					}
                                    }
				else {
				    if(Event.xbutton.x < 15) cdcontrol(DNTRACK);
				    }
				break;			 
			 case DestroyNotify:	/* Destroy window */
				XFreeGC(Disp, WinGC);
				XDestroyWindow(Disp, Win);
				XDestroyWindow(Disp, Iconwin);
				XCloseDisplay(Disp);				
				exit(0); 
				break;			
			}
		
		XFlush(Disp);
		RedrawWindow(&currentXPM);
		}
	usleep(50000L); 
 	RedrawWindow(&currentXPM);	
	}
}

/****************************************************************************/
void GetXPM(void)
{
	XWindowAttributes Attributes;
	int Ret;

	
	XGetWindowAttributes(Disp,Root,&Attributes);		
	cdplayerXPM.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
	Ret = XpmCreatePixmapFromData(Disp, Root, cdplayer, &cdplayerXPM.pixmap, 
				      &cdplayerXPM.mask, &cdplayerXPM.attributes);

        ledXPM.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
        Ret = XpmCreatePixmapFromData(Disp, Root, led, &ledXPM.pixmap,
                                      &ledXPM.mask, &ledXPM.attributes);
      

	cdpausedXPM.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
        Ret = XpmCreatePixmapFromData(Disp, Root, cdpaused, &cdpausedXPM.pixmap,
                                      &cdpausedXPM.mask, &cdpausedXPM.attributes);    



        currentXPM=cdplayerXPM;	
	if(Ret != XpmSuccess)
        {
                fprintf(stderr, "ascd: not enough free color cells\n");
                exit(1);
        }          
}

/****************************************************************************/
int FlushExpose(Window w)
{
	XEvent dummy;
	int i=0;
	
	while (XCheckTypedWindowEvent (Disp, w, Expose, &dummy))i++;
	return i;
}

/****************************************************************************/
void RedrawWindow(XpmIcon *Icon)
{	
	int track[2];
	int cdtime[3];
	static time_t last_flash_time;
	static flash=0;

	if(playing) cd_status();
     /*   close(cd_fd); */

	if(playing) currenttrack=cur_track;

	if((cur_track==maxtracks-1) && 
           (playing) && (!paused) && (cur_cdmode != 1))
           cdcontrol(STOP);
        else
	   if(!paused && lasttime==cur_pos_rel) return;

	if(paused)
	{
	   time_t flash_time = time(NULL);
	   if (flash_time == last_flash_time) return;
	   last_flash_time = flash_time;
	   flash = !flash;
	} else {
	   /* reset these if not paused, so next pause goes smoothly */
	   last_flash_time = 0;
	   flash = 1;
	}

        track[1]= (currenttrack % 10);
        track[0]= (currenttrack / 10);     


	lasttime=cur_pos_rel;
	if(playing||paused)
	{
	 cdtime[0]=(cur_pos_rel / 60);
	 cdtime[1]=((cur_pos_rel % 60) / 10);
	 cdtime[2]=(cur_pos_rel % 10);
	} else {
         cdtime[0]=0;
	 cdtime[1]=0;
	 cdtime[2]=0;
        }

        if(playing) currentXPM=cdpausedXPM;
        else currentXPM=cdplayerXPM;
	FlushExpose(Win);
	FlushExpose(Iconwin);
	
	XShapeCombineMask(Disp, Win, ShapeBounding, 0, 0, 
			  currentXPM.mask, ShapeSet);
	XShapeCombineMask(Disp, Iconwin, ShapeBounding, 0, 0, 
			  currentXPM.mask, ShapeSet);

	XCopyArea(Disp,currentXPM.pixmap,Win,WinGC,
		  0,0,currentXPM.attributes.width, currentXPM.attributes.height,0,0);	
	XCopyArea(Disp,currentXPM.pixmap,Iconwin,WinGC,
       	          0,0,currentXPM.attributes.width, currentXPM.attributes.height,0,0);


        XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
                  9*track[0],0, 9,11, 28,36);

        XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
                  9*track[0],0, 9,11, 28,36);           


	XCopyArea(Disp, ledXPM.pixmap, Win, WinGC, 
                  9*track[1],0, 9,11, 37,36);

        XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
                  9*track[1],0, 9,11, 37,36);      


 
	if(!paused || flash)
        {
           XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
                  9*cdtime[0],0, 9,11, 2,2);

           XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
                  9*cdtime[0],0, 9,11, 2,2);               

           XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
                  9*10,0, 6,11, 11,2);

           XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
                  9*10,0, 4,11, 11,2);               

           XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
                  9*cdtime[1],0, 9,11, 15,2);

           XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
                  9*cdtime[1],0, 9,11, 15,2);    

           XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
                  9*cdtime[2],0, 9,11, 24,2);

           XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
                  9*cdtime[2],0, 9,11, 24,2);    
	}

}

/****************************************************************************/
Pixel GetColor(char *ColorName)
{
	XColor Color;
	XWindowAttributes Attributes;
	
	XGetWindowAttributes(Disp,Root,&Attributes);
	Color.pixel = 0;
	
	if (!XParseColor (Disp, Attributes.colormap, ColorName, &Color)) 
		fprintf(stderr,"ascd: can't parse %s\n", ColorName);
	else if(!XAllocColor (Disp, Attributes.colormap, &Color)) 
		fprintf(stderr,"ascd: can't allocate %s\n", ColorName);       
	
	return Color.pixel;
}
